package	jp.co.sds_corp.fixconnect.transport;

import	java.io.IOException;
import	java.net.InetAddress;
import	java.net.ServerSocket;
import	java.net.Socket;
import	java.security.GeneralSecurityException;
import	java.util.Vector;

import	javax.net.ServerSocketFactory;
import	javax.net.ssl.SSLContext;

import	jp.co.sds_corp.fixconnect.FIXException;
import	jp.co.sds_corp.fixconnect.message.FIXMessage;
import	jp.co.sds_corp.fixconnect.message.IMessageReceiver;
import	jp.co.sds_corp.fixconnect.message.MessageReceiveEvent;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class Acceptor extends AbstractTransporter
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	IConfiguration	config;
	ServerSocket	svrSock;
	Vector<FIXSession> vecSession = new Vector<FIXSession>();

	public Acceptor() {
		backlog = 5;
	}

	@Override
	public void init( IConfiguration config ) throws Exception {
		super.init( config );

		this.config = config;
		String var;
		var = config.getAttributes().get( "backlog", null );
		if ( var != null && var.length() > 0 ) {
			backlog = Integer.valueOf( var ).intValue();
		}

		setStopped( false );
	}

	public void start() throws FIXException {
		try {
			InetAddress addr;
			if ( getHost() != null ) {
				addr = InetAddress.getByName( host );
			} else {
				addr = InetAddress.getLocalHost();
			}
			logger.info( this, "bind host=[" + host + "]"
				+ " port=[" + port + "]" );
			if ( isSsl()) {
				svrSock = createSSLServerSocket( port, backlog, addr );
			} else {
				svrSock = new ServerSocket( port, backlog, addr );
			}
			accept();
		} catch ( Exception ex ) {
			if ( isStopped()) {
				return;
			}
			logger.error( this, "bind error", ex );
			throw( new FIXException( "bind error", ex ));
		}
	}

	protected ServerSocket createSSLServerSocket( int port, int backlog,
			InetAddress addr )
		throws IOException, GeneralSecurityException, Exception
	{
		logger.debug( this, "create SSLServerSocket" );
		SSLContext context = createSSLContext();
		ServerSocketFactory factory = context.getServerSocketFactory();
		ServerSocket ret = factory.createServerSocket( port, backlog, addr );
		return ret;
	}

	public boolean isBound() {
		if ( svrSock == null ) {
			return false;
		}
		return svrSock.isBound();
	}

	protected void accept() throws FIXException {
		for ( ;; ) {
			try {
				Socket sock = svrSock.accept();
				FIXSession session = new FIXSession( sock );
				session.init( config );
				session.addMessageReceiver( this );
				session.addEventListener( this );
				vecSession.add( session );

				FIXSocket socket = session.getSocket();
				socket.setRecvBufSize( getRecvBufSize() );
				socket.setSocketTimeout( getSocketTimeout() );

				TransportEvent ev = new TransportEvent( session,
											TransportEvent.Type.ACCEPT );
				fireEventOccurred( ev );
				SessionState oldState = session.getState();
				session.setState( SessionState.CONNECTED );
				fireStateChanged( session, oldState, SessionState.CONNECTED );

				LogonManager logMan = new LogonManager( session );
				session.addMessageReceiver( "A", logMan );
				session.addMessageReceiver( "5", logMan );

				logger.info( this, "start session. id=" + session.getSessionId()
						+ " from=" + session.getSocket().getRemoteAddress() );
				session.start();
			} catch ( Exception ex ) {
				throw( new FIXException( "accept error", ex ));
			}
		}
	}

	public class LogonManager implements IMessageReceiver
	{
		FIXSession session;
		public LogonManager( FIXSession session ) {
			this.session = session;
		}

		public void onMessage( MessageReceiveEvent ev ) {
			FIXMessage req = ev.getMessage();
			try {
				if ( req.getMsgType().equals( "A" )) {
					logon( req );
				} else
				if ( req.getMsgType().equals( "5" )) {
					logoout( req );
				}
			} catch ( Exception ex ) {
				logger.error( this, "response send error", ex );
			}
		}

		public void logon( FIXMessage req ) throws FIXException {
			FIXMessage res = session.getParser().createMessage( "A" );
			res.setField( "EncryptMethod", req.getField( "EncryptMethod" ));
			res.setField( "HeartBtInt", req.getField( "HeartBtInt" ));
			session.send( res );
			session.setState( SessionState.LOGONED );
			logger.info( this, "logoned." + session );
		}

		public void logoout( FIXMessage req ) throws FIXException {
			FIXMessage res = session.getParser().createMessage( "5" );
			session.send( res );
			session.setState( SessionState.CONNECTED );
			logger.info( this, "logouted." + session );
		}
	}

	public void stop() {
		setStopped( true );
		try {
			svrSock.close();
		} catch ( Exception ex ) {
			logger.error( this, "server socket close error.", ex );
		}
		for ( int i = 0; i < vecSession.size(); i++ ) {
			FIXSession session = vecSession.get( i );
			logger.info( this, "close session." + session );
			try {
				session.close();
			} catch ( Exception ex ) {
				logger.error( this, "socket close error.", ex );
			}
			logger.info( this, "session closed." + session );
		}
	}

	int	backlog;	// キュー長
	boolean	stopped;	// 終了フラグ

	/**
	* キュー長を設定します。
	* @param	var		設定するデータ
	*/
	public void setBacklog( int var ) {
		this.backlog = var;
	}
	/**
	* キュー長を取得します。
	* @return	キュー長
	*/
	public int getBacklog() {
		return backlog;
	}

	/**
	* 終了フラグを設定します。
	* @param	var		設定するデータ
	*/
	public synchronized void setStopped( boolean var ) {
		this.stopped = var;
	}
	/**
	* 終了フラグを取得します。
	* @return	終了フラグ
	*/
	public synchronized boolean isStopped() {
		return stopped;
	}

}
