package	jp.co.sds_corp.fixconnect.transport;

import	java.io.IOException;
import	java.net.Socket;
import	java.security.GeneralSecurityException;
import	java.util.Vector;
import	java.util.concurrent.BlockingQueue;
import	java.util.concurrent.LinkedBlockingQueue;
import	java.util.concurrent.TimeUnit;

import	javax.net.SocketFactory;
import	javax.net.ssl.SSLContext;

import jp.co.sds_corp.fixconnect.FIXException;
import jp.co.sds_corp.fixconnect.message.FIXMessage;
import jp.co.sds_corp.fixconnect.message.IMessageReceiver;
import jp.co.sds_corp.fixconnect.message.MessageReceiveEvent;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class Initiator extends AbstractTransporter
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	IConfiguration config;
	LogonManager manager;

	public Initiator() {
	}

	@Override
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		FIXSession session = new FIXSession();
		session.init( config );
		setSession( session );

		session.addMessageReceiver( this );
		session.addEventListener( this );
		session.addStateChangeListener( this );

		manager = new LogonManager( session );
		session.addMessageReceiver( "A", manager );
		session.addMessageReceiver( "5", manager );
		session.addMessageReceiver( "3", manager );
		session.addMessageReceiver( "j", manager );
		session.addStateChangeListener( manager );

		this.config = config;
	}

	public void start() {
		try {
			logger.debug( this, "connect to host=[" + host + "]"
				+ " port=[" + port + "]" );

			Socket sock;
			if ( isSsl()) {
				sock = createSSLSocket( host, port );
			} else {
				sock = new Socket( host, port );
			}
			FIXSocket socket = new FIXSocket( sock );
			socket.setRecvBufSize( getRecvBufSize() );
			socket.setSocketTimeout( getSocketTimeout() );
			getSession().setSocket( socket );

			session.start();
		} catch ( Exception ex ) {
			logger.error( this, "connect error", ex );
		}
	}

	protected Socket createSSLSocket( String host, int port )
		throws IOException, GeneralSecurityException, Exception
	{
		logger.debug( this, "create SSLSocket" );
		SSLContext context = createSSLContext();
		SocketFactory factory = context.getSocketFactory();
		Socket ret = factory.createSocket( host, port );
		return ret;
	}

	public class LogonManager implements IMessageReceiver, IStateChangeListener
	{
		FIXSession session;
		Vector<BlockingQueue<Boolean>> listQueue =
			new Vector<BlockingQueue<Boolean>>();

		public LogonManager( FIXSession session ) {
			this.session = session;
		}

		public void onMessage( MessageReceiveEvent ev ) {
			FIXMessage res = ev.getMessage();
			try {
				if ( res.getMsgType().equals( "A" )) {
					logoned();
				} else
				if ( res.getMsgType().equals( "5" )) {
					logouted();
				}
			} catch ( Exception ex ) {
				logger.error( this, "response receive error", ex );
			}
		}

		void addQueue( BlockingQueue<Boolean> queue ) {
			listQueue.add( queue );
		}

		void removeQueue( BlockingQueue<Boolean> queue ) {
			listQueue.remove( queue );
		}

		public void logoned() {
			for ( BlockingQueue<Boolean> queue : listQueue ) {
				queue.offer( true );
			}
		}

		public void logouted() {
			for ( BlockingQueue<Boolean> queue : listQueue ) {
				queue.offer( false );
			}
		}

		public void stateChanged( StateChangeEvent ev ) {
			switch ( ev.getNewState()) {
			case IDLE:
				logouted();
				break;
			}
		}
	}

	public void stop() {
		try {
			close();
		} catch ( Exception ex ) {
			logger.error( this, "close error", ex );
		}
	}

	public void close() throws FIXException {
		getSession().close();
	}

	public void onMessage( MessageReceiveEvent ev ) {
		super.onMessage( ev );
	}

	public synchronized boolean logon() throws FIXException {
		SessionState state = session.getState();
		if ( state != SessionState.CONNECTED ) {
			throw( new IllegalStateException( "state=" + state ));
		}
		BlockingQueue<Boolean> queue = new LinkedBlockingQueue<Boolean>();
		manager.addQueue( queue );

		FIXMessage req = new FIXMessage( "A" );
		req.setField( "EncryptMethod", "0" );
		req.setField( "HeartBtInt", "10" );
		req.setField( "ResetSeqNumFlag", "Y" );
		if ( getSession().getTransportProtocol() != null ) {
			String ver = getApplVersion();
			if ( ver != null ) {
				req.setField( "DefaultApplVerID", getApplVersion() );
			}
		}
		getSession().send( req );
		getSession().setState( SessionState.CONNECTED, SessionState.LOGON_TRYING );

		Boolean ret = null;
		try {
			ret = queue.poll( 10, TimeUnit.SECONDS );
		} catch ( InterruptedException iex ) {
		} finally {
			manager.removeQueue( queue );
		}
		logger.debug( this, "logon result=" + ret );
		if ( ret == null ) {
			session.setState( SessionState.LOGON_TRYING, SessionState.CONNECTED );
			TimeoutException ex = new TimeoutException( "can't receive logon response" );
			// logger.error( this, "logon error", ex );
			throw( ex );
		}
		if ( ret ) {
			session.setState( SessionState.LOGON_TRYING, SessionState.LOGONED );
		} else {
			session.setState( SessionState.LOGON_TRYING, SessionState.CONNECTED );
		}
		return ret.booleanValue();
	}

	static final String[][] arrVer = {
		{ "FIX.2.7", "0" },
		{ "FIX.3.0", "1" },
		{ "FIX.4.0", "2" },
		{ "FIX.4.1", "3" },
		{ "FIX.4.2", "4" },
		{ "FIX.4.3", "5" },
		{ "FIX.4.4", "6" },
		{ "FIX.5.0", "7" },
		{ "FIX.5.0SP1", "8" },
		{ "FIX.5.0SP2", "9" }
	};
	private String getApplVersion() {
		String ret = null;
		String ver = getSession().getVersion();
		for ( String[] def : arrVer ) {
			if ( def[0].equals( ver )) {
				ret = def[1];
				break;
			}
		}
		return ret;
	}

	public synchronized boolean logout() throws FIXException {
		SessionState state = session.getState();
		if ( state != SessionState.LOGONED ) {
			throw( new IllegalStateException( "state=" + state ));
		}
		BlockingQueue<Boolean> queue = new LinkedBlockingQueue<Boolean>();
		manager.addQueue( queue );

		FIXMessage req = new FIXMessage( "5" );
		getSession().send( req );

		Boolean ret = null;
		try {
			ret = queue.poll( 10, TimeUnit.SECONDS );
		} catch ( InterruptedException ex ) {
		} finally {
			manager.removeQueue( queue );
		}
		logger.debug( this, "logout result=" + ret );
		if ( ret == null ) {
			TimeoutException ex = new TimeoutException( "can't receive logout response" );
			// logger.error( this, "logout error", ex );
			throw( ex );
		}
		session.setState( SessionState.LOGONED, SessionState.CONNECTED );
		return ret.booleanValue();
	}

	public void send( FIXMessage msg ) throws FIXException {
		getSession().send( msg );
	}

	boolean waitState( SessionState current, SessionState expected ) {
		boolean ret = false;
		for ( ;; ) {
			if ( getSession().getState() == expected ) {
				ret = true;
				break;
			}
			if ( getSession().getState() != current ) {
				break;
			}
			try {
				Thread.sleep( 100 );
			} catch ( InterruptedException ex ) {
			}
		}
		return ret;
	}

	FIXSession	session;	// セッション情報

	/**
	* セッション情報を設定します。
	* @param	var		設定するデータ
	*/
	public synchronized void setSession( FIXSession var ) {
		this.session = var;
	}
	/**
	* セッション情報を取得します。
	* @return	session
	*/
	public synchronized FIXSession getSession() {
		return session;
	}

}
